import h5py
import hdf5storage
import numpy

class TrxMat:
    """
    Read `trx.mat` files generated by JBM's behaviour tagging pipeline.
    """

    def __init__(self, path):
        self.path = str(path)
        self._records = None
        self._memoized = {}

    @property
    def records(self):
        """
        list: Available record names.
        """
        if self._records is None:
            self._records = [] # fallback
            trx = h5py.File(self.path, 'r')
            try:
                self._records = list(trx['trx'].keys())
            finally:
                trx.close()
        return self._records

    @records.setter
    def records(self, records):
        if records is None:
            self._records = None
        else:
            raise ArgumentError("property `records` is read-only; can only be reset with value `None`")

    def __nonzero__(self):
        return bool(self.records)

    def __iter__(self):
        return iter(self.records)

    def __getitem__(self, record):
        data = self.read([record], lowlevel=False)
        if record not in data:
            levels = set(("id", "numero_larva_num"))
            records_to_zip = {record: data[record] \
                    for record in data.keys() if record not in levels}
            data = self.zip_records(records_to_zip)
            assert len(data) == 1
            record = next(iter(data.keys()))
        return data[record]

    def read(self, records=[], lowlevel=False):
        """
        Read variables from file.

        Variable names to be read can be passed as argument `records`.
        An empty record list is interpreted as all available records.

        Returns a dictionary of (record name, data array) pairs (key, value).

        With `lowlevel=True`, the data array for each record is a (single-item)
        list of (single-item) tuples of arrays of arrays of floats.

        With `lowlevel=False`, the data array for each record is a multi-level
        dictionary with run identifiers (`str` *date_time*) as first-level keys,
        larva identifiers (`int`) as second-level keys, and series of data items
        as values, typically as `numpy.ndarray`.

        In addition, with `lowlevel=False`, keys *spine* and *contour*/*outline*
        are translated into *x_spine* and *y_spine*, or *x_contour* and
        *y_contour*, respectively.

        Last but not least, with `lowlevel=False`, records other then *id* and
        *numero_larva_num* are memoized.
        """
        if records:
            if isinstance(records, str):
                records = records.split()
            if not lowlevel:
                records, memoized_records = self._parse_record_names(records)
            file = h5py.File(self.path, 'r')
            try:
                trx = {}
                for record in records:
                    varname = 'trx/'+record
                    refs = file[varname][0,:]
                    # np arrays (and transpose) for backward compatibility
                    trx[varname] = numpy.empty(refs.shape, dtype=object)
                    if record == 'id':
                        for i, ref in enumerate(refs):
                            trx[varname][i] = numpy.array([file[ref][:].tobytes().decode('utf-16')])
                    else:
                        for i, ref in enumerate(refs):
                            trx[varname][i] = numpy.transpose(file[ref][:])
                    assert not numpy.isscalar(trx[varname])
            finally:
                file.close()
        elif lowlevel:
            # TODO: replace hdf5storage or deprecate this use case
            trx = hdf5storage.loadmat(self.path)
        else:
            # explicit record names required for memoization
            return self.read(self.records)
        data = {}
        for varname in trx:
            record = varname[4:]
            vardata = trx[varname]
            if numpy.isscalar(vardata): # hdf5storage
                vardata = vardata[0]
            try:
                hook = getattr(self, record+"_hook")
            except AttributeError:
                pass
            else:
                vardata = hook(vardata)
            data[record] = vardata
        if not lowlevel:
            data = self._multilevel_data(data, memoized_records)
        return data

    def _parse_record_names(self, records):
        records = set(records)
        for special, translation in (("outline", "contour"),):
            if special in records:
                records.remove(special)
                records.add(translation)
        for special in ("spine", "contour"):
            if special in records:
                records.remove(special)
                records |= set(("x_"+special, "y_"+special))
        levels = set(("id", "numero_larva_num"))
        if records - levels:
            records |= levels
            return self._filter_memoized(records)
        else:
            return records, set()

    def _filter_memoized(self, records):
        new_records = records - set(self._memoized.keys())
        memoized_records = records - new_records
        return new_records, memoized_records

    def _flatten_record(self, data, astype=None):
        data = [array.squeeze()[...] for array in data]
        if astype is str:
            data = [array.tolist() for array in data]
        elif astype is not None:
            data = [array.astype(astype) for array in data]
        return data

    def _multilevel_data(self, data, memoized_records=None):
        levels = set(("id", "numero_larva_num"))
        records = set(data.keys()) - levels
        if not records and not memoized_records:
            return data
        #
        restructured_data = {record: {} for record in records}
        try:
            runs = self._memoized["id"]
        except KeyError:
            self._memoized["id"] = runs = self._flatten_record(data["id"], str)
        try:
            larvae = self._memoized["numero_larva_num"]
        except KeyError:
            larvae = numpy.array(self._flatten_record(data["numero_larva_num"], int))
            self._memoized["numero_larva_num"] = larvae
        #
        records = {record_name: numpy.array(
                    self._flatten_record(data[record_name]),
                    dtype=object) for record_name in records}
        for run in set(runs):
            run_ = numpy.array([run_ == run for run_ in runs])
            larvae_ = larvae[run_]
            for larva in set(larvae_):
                larva_ = larvae_ == larva
                for record in records:
                    data__ = records[record][run_][larva_].tolist()
                    assert len(data__) == 1
                    data__ = data__[0]
                    data_ = restructured_data[record].get(run, {})
                    data_[larva] = data__
                    restructured_data[record][run] = data_
        #
        self._memoized.update(restructured_data)
        if memoized_records:
            for record in memoized_records:
                restructured_data[record] = self._memoized[record]
        return restructured_data

    def zip_records(self, records):
        if len(records) != 2:
            raise NotImplementedError
        it = iter(records.items())
        xname, x = next(it)
        yname, y = next(it)
        coord, name = xname.split('_')
        assert coord in 'xy'
        if coord == 'y':
            xname, x, yname, y = yname, y, xname, x
        assert name == yname.split('_')[1]
        zipped = {name: x}
        for run in x:
            for larva in x[run]:
                xy = []
                for xt, yt in zip(x[run][larva], y[run][larva]):
                    xyt = numpy.stack((xt, yt), axis=1)
                    xy.append(xyt)
                x[run][larva] = xy
        return zipped
